#! /bin/sh
# $Id: gplize.sh 1772 2007-07-16 22:46:42Z sigoure $
#<GPL>
# Add the GPL copyright notice to sources. 
# Copyright (C) 2006  SIGOURE Benoit
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
# USA.
#</GPL>

usage()
{
  cat >&2 <<EOF
Usage: gplize.sh [OPTIONS...] <file> [files...]
  -a, --auto             Auto mode. In this mode the script will be
                         non-interactive. If the script hasn't got enough
                         information, it won't prompt the user, instead it
                         will simply exit 1. This entails that you probably
                         must pass the --copyright and --project options.
                         This option also turns on --assume-yes.
                         When GPLizing new files, the comment style to use
                         will be chosen depending on the file's extension.
  -c, --copyright NAME   Default copyright holder.
  -d, --debug            Enable debugging.
  -h, --help             Show this message.
  -l, --license FILE     Instead of GPL, use the license in FILE. The FILE may
                         contain any number of the following placeholders:
                           <DESC>: Description of the file.
                           <YEARS>: Years of copyright.
                           <OWNER>: Copyright holder.
                           <FILE>: Path of the file GPLized. (eg: ../foo/bar.c)
                           <BNAME>: Basename of the file GPLized. (eg: bar.c)
  -n, --preview          Print a preview before GPLizing a file.
  -m, --postcmd CMD      Performs the command CMD whenever a new file is
                         GPLized. The CMD can contain the same placeholders as
                         described for the --license option, eg:
                             --postcmd 'svn propset svn:keywords Id <FILE>'
                         Will enable the svn keyword \$Id\$ on GPLized files.
  -p, --project NAME     Default project name.
  -u, --update           Update the project name. (requires -p | --project)
  -y, --assume-yes       Assume yes to all questions.

Warning: This script will not preserve the setuid/setgid/etc bits on file
         modified. Only the executable mode will be preserved.
Copyright (C) 2006  SIGOURE Benoit
This is free software; see the source for copying conditions.  There is NO
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
EOF
  exit 1
}

test $# -lt 1 && usage

# Usage: yesno <QUESTION>
yesno()
{
  echo -n "??? $1 ? [Y/n] "
  if [ $assume_yes = yes ]; then
    echo "y (--assume-yes)"
    return 0
  fi

  read answer || return 1
  case "$answer" in
    n* | N* ) return 1 ;;
    *) return 0 ;;
  esac
  return 1 # Should never be here..
}

# Usage: gplize <FILE>
gplize()
{
  bname=`basename $1`

  custom_txt=''
  test x$custom_license = xyes && \
    custom_txt=" with your custom license in $license_file"
  yesno "Do you want to GPLize $1$custom_txt" || return 1

  echo '1. Shell (#)'
  echo '2. C/C++/Stratego (/* * */)'
  echo '3. Caml (* ... *)'
  echo '4. Other'
  echo -n '??? Comment style [1/2/3/4] '
  if [ x$auto_mode = xno ]; then
    read answer || return 1
    case "$answer" in
      1* ) c_start='#';  c_mid='#';  c_end=''    ;;
      2* ) c_start='/*'; c_mid=' *'; c_end=' */' ;;
      3* ) c_start='(*'; c_mid=' *'; c_end=' *)' ;;
      4* ) c_start='';   c_mid='';   c_end=''
        reuse=no
        test x$prev_comment_style = xyes && \
          yesno "Re-use: $prev_c_start, $prev_c_mid, $prev_c_end" && \
          c_start="$prev_c_start" && \
          c_mid="$prev_c_mid" &&     \
          c_end=$prev_c_end &&       \
          reuse=yes
        if test x$reuse = xno; then
          echo -n 'Type how comments start: '
          read c_start || return 1
          echo -n 'Type what should be placed at the beginning of each line: '
          read c_mid || return 1
          echo -n 'Type how comments end: '
          read c_end || return 1
          prev_c_start="$c_start"; prev_c_mid="$c_mid"; prev_c_end=$c_end
          prev_comment_style=yes
        fi
        ;;
      *) echo "Skipping $1"; return 1;;
    esac
  else # auto_mode = yes.
    case "$bname" in # Guess the comment's style to use.
      *.c | *.cc | *.h | *.hh | *.C | *.cpp | *.cc | *.cxx | *.hxx | *.hcc \
      | *.hpp | *.tcc | *.str)
        echo '(auto) 2'; c_start='/*'; c_mid=' *'; c_end=' */' ;;
      *.ml | *.ml[ily])
        echo '(auto) 3'; c_start='(*'; c_mid=' *'; c_end=' *)' ;;
      *.sh | *.sh.in | bootstrap* | *.m4 | configure* | *.pl | *.pm | *.rb \
      | Makefile* | *.py)
        echo '(auto) 1'; c_start='#';  c_mid='#'; c_end='' ;;
      *.sdf) echo '(auto) SDF'; c_start='%%';  c_mid='%%'; c_end='' ;;
      *) echo "Can't guess the comment style to use in $1" >&2; exit 1;;
    esac
  fi
  echo ">>> GPLizing $1 ... "

  if [ x$auto_mode = xno ]; then
    # Copyright descr
    prev_descr="${prev_descr-\$1: does something for ProjectName}"
    echo "Give one line to with the program's name and a brief idea of\
 what it does."
    test "x${prev_descr-set}" != xset && echo_descr="Default: $prev_descr "
    echo "$echo_descr(\$1 will be replaced by $bname)"
    echo -n "> "
    read descr || return 1
    test x"$descr" = x && descr="$prev_descr"
    test x"$descr" = x && \
    {
      echo 'You must give a copyright description' >&2
      return 1
    }
    prev_descr="$descr"
    descr=`echo "$descr" | sed "s/\\\$1/$bname/g"`
    descr=`echo "$descr" | sed '/\.$/!s/$/./'` # Add trailing `.' if necessary
  else # auto_mode = yes -> no description.
    descr=''; prev_descr=''
  fi

  # Part of a project?
  echo "If this file is part of a project, enter its name, otherwise simply\
 hit return"
  test x"$prev_proj" = x && prev_proj="$default_project"
  test x"$prev_proj" != x && echo "Default: $prev_proj"
  echo -n "> "
  if [ x$auto_mode = xno ]; then
    read proj || return 1
  else
    echo '(auto mode: skipped)'
    proj="$prev_proj"
  fi
  test x"$proj" = x && proj="$prev_proj"
  prev_proj="$proj"
  if [ x"$proj" = x ]; then
    proj="This program"
    sed_descr="$descr"
  else
    sed_descr="$descr\\
$c_mid This file is part of $proj."
    descr="$descr
$c_mid This file is part of $proj."
  fi

  # Copyright holder
  test "x${prev_holder-set}" = xset && prev_holder="$default_copyright"
  test "x${prev_holder-set}" != xset && echo_holder="[$prev_holder] "
  echo -n "Who is the copyright holder $echo_holder? "
  if [ x$auto_mode = xno ]; then
    read holder || return 1
  else
    echo '(auto mode: skipped)'
    holder="$prev_holder"
  fi
  test x"$holder" = x && holder="$prev_holder"
  test x"$holder" = x && \
  {
    echo 'You must name a copyright holder' >&2
    return 1
  }
  holder=`echo "$holder" | sed '/\.$/!s/$/./'` # Add trailing `.' if necessary
  prev_holder="$holder"

  # Years of copyright
  # FIXME: the following will miss out added files not yet commited.
  if svn diff $1 >/dev/null 2>/dev/null; then
    years=`svn log $1 2>/dev/null \
           | awk '/^r[0-9]/ {split($0, f, "|"); print substr(f[3], 2, 4)}' \
           | sort -nu | xargs | sed 's/ /, /g'`
    test x"$years" = x && years=`date '+%Y'` # File just added.
    echo "Adding copyright for the following years: $years"
    test x$auto_mode = xno && yesno "Do you want to add more years" && \
    {
      echo -n "Extra years: "
      read extra || return 1
      test x"$extra" != x && years="$years, $extra"
    }
  elif [ x"`git log --pretty=format:%h $1`" != x ]; then
    years=`git log --pretty=format:%aD main.cc | awk '{ print $4 }' \
           | sort -nu | xargs | sed 's/ /, /g'`
    echo "Adding copyright for the following years: $years"
    test x$auto_mode = xno && yesno "Do you want to add more years" && \
    {
      echo -n "Extra years: "
      read extra || return 1
      test x"$extra" != x && years="$years, $extra"
    }
  else
    if [ x$auto_mode = xyes ]; then
      echo "The file \`$1' is not under version control, can't guess the\
 copyright years." >&2
      exit 1
    fi
    test "x${prev_years-set}" != xset && echo_years="[$prev_years] "
    echo -n "Add the copyright for which years $echo_years? "
    read years || return 1
    test x"$years" = x && years="$prev_years"
    test x"$years" = x && \
    {
      echo 'You must specify at least one year' >&2
      return 1
    }
    prev_years="$years"
  fi

  # When using custom licenses, we must escape @s for sed
  sed_descr=`echo "$sed_descr"   | sed 's/@/\\\\@/g'`
  sed_years=`echo "$years"   | sed 's/@/\\\\@/g'`
  sed_holder=`echo "$holder" | sed 's/@/\\\\@/g'`
  sed_file=`echo "$1"        | sed 's/@/\\\\@/g'`
  # We will use $c_end with the 'a' command in sed. When it starts with
  # spaces, they are ignored by BSD sed unless escaped.
  sed_c_end=`echo "$c_end"   | sed '/^  */ s/^/\\\\/'`

  # Preview?
  if [ $preview = yes ]; then
    echo "Preview of the copyright notice to be added:"
    if [ x$custom_license = xno ]; then # GPL
      echo "$c_start $descr"
      echo "$c_mid Copyright (C) $years  $holder"
      awk '/^#<GPL>$/{r=1;next}/^#<\/GPL>$/{r=0}r==1{$1="";print}' "$0" |\
        sed "1d;2d;s/^/$c_mid/;4s/This program/$proj/"
      echo "$c_end"
    else
      sed "1s@^@$c_start @; \$a\\
$sed_c_end
           1!s@^@$c_mid @;
           s@<DESC>@$sed_descr@g;
           s@<YEARS>@$sed_years@g;
           s@<OWNER>@$sed_holder@g;
           s@<FILE>@$sed_file@g;
           s/<BNAME>/$bname/g;
           s/ *$//" "$license_file" || return 1
    fi
    yesno "Add the copyright notice" || return 1
  fi

  # Let's rock
  if [ x"`sed '1s/^\(..\).*/\1/;q' $1`" = 'x#!' ]; then
    shebang=yes
    echo "Shebang detected"
    # sed 'q' is a faster equivalant to sed '1!d'
    sed 'q' $1 >$1.gplize$$
    test x$custom_license = xno && echo "$c_start $descr" >>$1.gplize$$
  else
    shebang=no
    test x$custom_license = xno && echo "$c_start $descr" >$1.gplize$$
  fi
  # Preserve the executable bit (others are lost -- setuid etc...)
  test -x "$1" && chmod a+x $1.gplize$$
  if test x$custom_license = xno; then # GPL.
    echo "$c_mid Copyright (C) $years  $holder" >>$1.gplize$$
    awk '/^#<GPL>$/{r=1;next}/^#<\/GPL>$/{r=0}r==1{$1="";print}' "$0" |\
      sed "1d;2d;s/^/$c_mid/;4s/This program/$proj/" >>$1.gplize$$
    echo "$c_end" >>$1.gplize$$
  else # Custom license.
      sed "1s@^@$c_start @; \$a\\
$sed_c_end
           1!s@^@$c_mid @;
           s@<DESC>@$sed_descr@g;
           s@<YEARS>@$sed_years@g;
           s@<OWNER>@$sed_holder@g;
           s@<FILE>@$sed_file@g;
           s/<BNAME>/$bname/g;
           s/ *$//" "$license_file" >>$1.gplize$$ || return 1
  fi
  # Append the old content of the file in the GPLized version.
  if [ $shebang = yes ]; then
    sed '1d' $1 >>$1.gplize$$
  else
    cat $1 >>$1.gplize$$
  fi
  mv $1.gplize$$ $1 || {
    echo "Failed to GPLize $1" >&2
    return 1
  }

  # Perform the postcmd (if any)
  if [ x"$postcmd" != x ]; then
    _postcmd=`echo "$postcmd" | sed "s@<DESC>@$sed_descr@g;
           s@<YEARS>@$sed_years@g;
           s@<OWNER>@$sed_holder@g;
           s@<FILE>@$sed_file@g;
           s/<BNAME>/$bname/g"` || return 1
    eval "echo \"Executing post command: $_postcmd\""
    eval "$_postcmd"
  fi

  echo ">>> $1 GPLized"
}

# Default options
default_copyright=''
preview=no
default_project=''
update_project_name=no
assume_yes=no
custom_license=no
license_file=''
auto_mode=no
postcmd=''

# Parse arguments
while [ $# -gt 0 ]; do
  case "$1" in
    --*=*)
      opt=`echo "$1" | sed 's/=.*//'`
      val=`echo "$1" | sed 's/[^=]*=//'`
      shift
      set dummy "$opt" "$val" ${1+"$@"}
      shift
      ;;
  esac

  case "$1" in
    -a | --auto)      auto_mode=yes; assume_yes=yes;;
    -c | --copyright) shift; default_copyright="$1";;
    -d | --debug)     set -x
                      ;;
    -h | --help)      usage;;
    -l | --license)   custom_license=yes; shift; license_file="$1";;
    -n | --preview)   preview=yes;;
    -m | --postcmd)   shift; postcmd="$1";;
    -p | --project)   shift; default_project="$1";;
    -u | --update)    update_project_name=yes;;
    -y | --assume-yes | --yes) assume_yes=yes;;
    -*) echo "$0: Unknown argument \`$1'. Use $0 --help." >&2; shift;;
    *) break;;
  esac
  shift
done

# Enforce a correct value for preview if auto_mode is enabled.
test x$auto_mode = xyes && preview=no

if [ $update_project_name = yes ] && [ x"$default_project" = x ]; then
  update_project_name=no
fi

test $# -eq 0 && echo 'gplize: No input file.' && usage

while [ $# -gt 0 ]; do
  test -f $1 || { shift; continue; }
  echo ">>> $1"
  grep -q Copyright $1 || {
    echo "No copyright found in $1" >&2
    gplize $1 || echo "!!! Failed to GPLize $1" >&2
    shift
    continue
  }

  if grep -q '59 Temple Place' $1; then
    echo ">>> Updating the address of the FSF (changed since april 2005)"
    # sed -i isn't portable
    sed "s/59 Temple Place.*Suite 330/51 Franklin Street, Fifth Floor/; \
         s/02111-1307/02110-1301/;" $1 >$1.gplize$$
    # Preserve the executable bit (others are lost -- setuid etc...)
    test -x "$1" && chmod a+x $1.gplize$$
    mv $1.gplize$$ $1 || { shift; continue; }
  fi
  
  file_years=`awk \
  '/Copyright/{gsub(/[^0-9,]/,"");split($0,a,",");for(i in a){print a[i]}}' $1 \
  | sort -nu | xargs | sed 's/ /, /g'`
  echo "Years detected in the $1: $file_years"

  svn diff $1 >/dev/null 2>/dev/null \
  || test x"`git log --pretty=format:%h $1`" != x \
  || {
    echo "$1 is not under SVN/Git control" >&2
    shift
    continue
  }

  if svn diff $1 >/dev/null 2>/dev/null; then
    svn_years=`svn log $1 2>/dev/null \
               | awk '/^r[0-9]/ {print substr($5, 1, 4)}' \
               | sort -nu | xargs | sed 's/ /, /g'`
    test x"$svn_years" = x && svn_years=`date '+%Y'` # File just added.
    echo "svn says: $svn_years"
  elif [ x"`git log --pretty=format:%h $1`" != x ]; then
    svn_years=`git log --pretty=format:%aD main.cc | awk '{ print $4 }' \
               | sort -nu | xargs | sed 's/ /, /g'`
    test x"$svn_years" = x && svn_years=`date '+%Y'` # File just added.
    echo "git says: $svn_years"
  else
    echo internal error
    exit 42
  fi

  years=`echo "$file_years,$svn_years" | sed 's/ //g;s/[^0-9,]//g'`
  years=`awk "BEGIN{s=\"$years\"; split(s,a,\",\"); for(i in a)y[a[i]]=1;\
              for(i in y)print i}" | sort -nu | xargs | sed 's/ /, /g'`
  echo "Updating the copyright notice with years: $years"
  # sed -i isn't portable
  # FIXME: The following substitution is too greedy and will eat part of the
  # copyright holder if there's a digit in its name.
  sed "s/Copyright.*[0-9] /Copyright (C) $years /" $1 >$1.gplize$$
  # Preserve the executable bit (others are lost -- setuid etc...)
  test -x "$1" && chmod a+x $1.gplize$$
  mv $1.gplize$$ $1

  if [ $update_project_name = yes ]; then
    sed "/This program is free software/s/This program/$default_project/" \
      $1 >$1.gplize$$
    # Preserve the executable bit (others are lost -- setuid etc...)
    test -x "$1" && chmod a+x $1.gplize$$
    mv $1.gplize$$ $1
  fi

  shift
done
